from enum import Enum
import logging
import math
from PIL import Image, ImageDraw, ImageFont
from skit._types import Color, Alignment, Scale


logger = logging.getLogger(__file__)


# some defaults for fallback
try:
    _DEFAULT_FONT = ImageFont.truetype('Helvetica', 16)
except OSError:
    _DEFAULT_FONT = ImageFont.load_default()
_DEFAULT_COLOR = 'black'
_DEFAULT_THICKNESS = 1


class DrawCommand(Enum):
    TEXT = 'text'
    RECTANGLE = 'rect'
    IMAGE = 'image'


class SingleImageRenderer:
    def __init__(self, layouts):
        self._layouts = layouts

    def render(self, width: int, height: int, background: Color, commands: list[dict]) -> Image.Image:
        with Image.new('RGBA', (width, height), background) as im:
            d = ImageDraw.Draw(im)

            for cmd in commands:
                cmd = cmd.copy()    # don't destroy the original command list
                op = cmd.pop('op')
                match op:
                    case DrawCommand.TEXT:
                        self._render_text(d, **cmd)
                    case DrawCommand.RECTANGLE:
                        self._render_rectangle(d, **cmd)
                    case DrawCommand.IMAGE:
                        self._render_image(im, **cmd)
                    case _:
                        raise ValueError(cmd)

            return im

    def _render_text(self, d, layout, text, color, font):
        logger.debug(f"rendering text '{text}' at {layout}")
        layout = self._layouts[layout]

        match layout['h_align']:
            case Alignment.BEGIN:
                anchor_h = 'l'
                x = layout['x']
            case Alignment.MIDDLE:
                anchor_h = 'm'
                x = layout['x'] + (layout['width'] // 2)
            case Alignment.END:
                anchor_h = 'r'
                x = layout['x'] + layout['width']
            case _:
                raise ValueError(f"h_align value '{layout['h_align']}' unrecognized")

        match layout['v_align']:
            case Alignment.BEGIN:
                y = layout['y']
                anchor_v = 'a'
            case Alignment.MIDDLE:
                y = layout['y'] + (layout['height'] // 2)
                anchor_v = 'm'
            case Alignment.END:
                y = layout['y'] + layout['height']
                anchor_v = 'd'
            case _:
                raise ValueError(f"v_align value '{layout['v_align']}' unrecognized")

        d.text(
            [x, y],
            text,
            fill=color if color else _DEFAULT_COLOR,
            font=font if font else _DEFAULT_FONT,
            anchor=f"{anchor_h}{anchor_v}",
        )
    
    def _render_rectangle(self, d, layout, color, thickness, filled):
        logger.debug(f"rendering rectangle on {layout}")
        layout = self._layouts[layout]
        d.rectangle(
            [
                layout['x'],
                layout['y'],
                layout['x'] + layout['width'],
                layout['y'] + layout['height'],
            ],
            fill=color if filled and color else None,
            outline=color if color else _DEFAULT_COLOR,
            width=thickness if thickness else _DEFAULT_THICKNESS,
        )

    def _render_image(self, im, layout, image):
        logger.debug(f"rendering image at {layout}")
        layout = self._layouts[layout]
        with Image.open(image) as art:
            # compute new image scale
            proposed_scale = self._pick_image_size(art.width, art.height, layout['width'], layout['height'])
            match layout['scale']:
                case Scale.FIT:
                    art = art.resize(proposed_scale)
                case Scale.UP:
                    if art.width < proposed_scale[0] or art.height < proposed_scale[1]:
                        art = art.resize(proposed_scale)
                case Scale.DOWN:
                    if art.width > proposed_scale[0] or art.height > proposed_scale[1]:
                        art = art.resize(proposed_scale)
                case Scale.NONE:
                    pass    # there is nothing to do
                case _:
                    raise ValueError(f"scale value '{layout['scale']}' unrecognized")

            match layout['h_align']:
                case Alignment.BEGIN:
                    left = layout['x']
                case Alignment.MIDDLE:
                    left = layout['x'] + (layout['width'] - art.width) // 2
                case Alignment.END:
                    left = layout['x'] + layout['width'] - art.width
                case _:
                    raise ValueError(f"h_align value '{layout['h_align']}' unrecognized")

            match layout['v_align']:
                case Alignment.BEGIN:
                    top = layout['y']
                case Alignment.MIDDLE:
                    top = layout['y'] + (layout['height'] - art.height) // 2
                case Alignment.END:
                    top = layout['y'] + layout['height'] - art.height
                case _:
                    raise ValueError(f"v_align value '{layout['v_align']}' unrecognized")

            im.alpha_composite(art, (left, top))

    def _pick_image_size(self, img_width, img_height, layout_width, layout_height):
        if img_width == layout_width and img_height == layout_height:
            return img_width, img_height
        
        # unpacking the logic below...
        #   ratio_width = image.width / layout.width
        #   ratio_height = image.height / layout.height
        #   if either ratio is > 1, we are shrinking, so pick the larger one to make sure we fit
        #   if both ratios are < 1, we are growing, so pick the one closer to 1 to maximally fill
        #   either way, we're picking the largest number to use as a scale factor

        scale_factor = max(img_width / layout_width, img_height / layout_height)
        new_width = math.floor(img_width / scale_factor)
        new_height = math.floor(img_height / scale_factor)
        
        return new_width, new_height



from collections.abc import MutableSequence, Sequence
from itertools import cycle
import logging
from pathlib import Path
from typing import Iterator, Self, Mapping, Callable, TypeVar
import warnings
from skit.card import Card, CardManipulation
from skit._types import Real, LayoutDef, Color, FreeTypeFont


logger = logging.getLogger(__file__)

class Deck(MutableSequence, CardManipulation):
    "A deck of one or more cards."
    def __init__(self, card_count: int = 1, width: int | None = None, height: int | None = None):
        """Create a new deck with `card_count` cards in it."""
        opts = {}
        if width: opts['width'] = width
        if height: opts['height'] = height
        self._cards: list[Card] = [Card(**opts) for _ in range(card_count)]
    
    #region Container convenience
    def __add__(self, other: Self) -> Self:
        if not issubclass(type(other), Deck):
            raise NotImplemented

        d = Deck(0)
        d._cards = self._cards + other._cards
        return d
    #endregion

    #region Card manipulation
    def background(self, color: Color):
        "Set the background color for all cards in this deck."
        logger.debug(f"Deck.background({color})")
        for card in self._cards:
            card.background(color)
    
    def layout(self, name: str, layoutdef: LayoutDef):
        "Add a layout to every card in this deck."
        logger.debug(f"Deck.layout({name}, ...)")
        for card in self._cards:
            card.layout(name, layoutdef)
    
    def layouts(self, names: Sequence[str], layoutdefs: Sequence[LayoutDef]):
        "Add multiple layouts to every card in this deck."
        logger.debug(f"Deck.layouts(sequence of layouts)")
        for card in self._cards:
            card.layouts(names, layoutdefs)

    def layouts_map(self, layouts: Mapping[str, LayoutDef]):
        "Add multiple layouts from a dictionary to every card in this deck."
        logger.debug(f"Deck.layouts_map(map of name->layouts)")
        for card in self._cards:
            card.layouts_map(layouts)

    def text(self, text: str, layout: str, font: FreeTypeFont | None = None, color: Color | None = None):
        "Add a text string to every card in this deck."
        logger.debug(f"Deck.text({text})")
        for card in self._cards:
            card.text(text, layout, font, color)
    
    def rectangle(self, layout: str, color: Color | None = None, thickness: Real | None = None):
        "Draw a rectangle on every card in this deck."
        logger.debug(f"Deck.rect({layout})")
        for card in self._cards:
            card.rectangle(layout, color, thickness)

    def filled_rectangle(self, layout: str, color: Color):
        "Draw a filled rectangle on every card in this deck."
        logger.debug(f"Deck.filled_rect({layout})")
        for card in self._cards:
            card.filled_rectangle(layout, color)

    def image(self, image: Path, layout: str):
        "Draw an image on every card in this deck."
        logger.debug(f"Deck.image({image})")
        for card in self._cards:
            card.image(image, layout)

    def render_png(self, filename: str):
        """
        Render every card in this deck as a PNG.

        You may use `{index}` as part of the filename to ensure each card gets
        a unique name. For example,

        ```python
        deck.render_png("card_{index}.png")
        ```
        """
        logger.debug(f"Deck.render_png({filename})")

        if '{index}' not in filename:
            warnings.warn("'{index}' isn't in the filename, so images may overwrite one another")

        for index, card in enumerate(self._cards):
            card.render_png(filename.format_map({
                'index': index,
            }))

    def render_pdf(self, filename: str, resolution: int, single_file=True):
        """
        Render every card in this deck as a PDF.

        If `single_file` is True, then make one big PDF. Otherwise,
        make a series of individual PDFs. If you make individual PDFs,
        you may use `{index}` as part of the filename to ensure each card gets
        a unique name. For example,

        ```python
        deck.render_pdf("card_{index}.pdf")
        ```
        """
        logger.debug(f"Deck.render_pdf({filename})")

        if single_file:
            self._render_single_pdf(filename, resolution)
        else:
            self._render_multiple_pdf(filename, resolution)

    def _render_single_pdf(self, filename: str, resolution: int):
        pages = []
        for card in self._cards:
            pages.append(card._get_rgb_image_for_pdf(resolution))
        
        first = pages.pop(0)
        first.save(
            filename,
            format='PDF',
            resolution=resolution,
            save_all=True,
            append_images=pages,
        )

    def _render_multiple_pdf(self, filename: str, resolution: int):
        if '{index}' not in filename:
            warnings.warn("'{index}' isn't in the filename, so images may overwrite one another")

        for index, card in enumerate(self._cards):
            card.render_pdf(
                filename.format_map({'index': index}),  
                resolution=resolution,
                single_file=True,
            )
    #endregion

    #region Card sequence manipulation
    T = TypeVar('T')

    def for_each_if(
        self,
        values: Sequence[T],
        test: Callable[[T], bool],
        do: Callable[[Card, T], None],
    ):
        """
        For each element in values, call `test(element)`. If it returns `True`,
        then call `do(corresponding_card, element)`. For example,

        ```
        # if there's a 'stats' key in the data, print stats in the right place
        deck.for_each_if(
            data,
            lambda elem: 'stats' in elem,
            lambda card, elem: card.text(elem['stats'], layout='stats')
        )
        ```
        """
        for index, elem in enumerate(values):
            if test(elem):
                do(self._cards[index], elem)

    def backgrounds(self, colors: Sequence[str]):
        """
        Add backgrounds to each card in this deck. `colors` may be scalar
        or a sequence, varying the background per-card.
        """
        if not issubclass(type(colors), Sequence) or issubclass(type(colors), str):
            colors = [colors]
        logger.debug(f"Deck.backgrounds(sequence of colors)")
        for card, color in zip(self._cards, cycle(colors)):
            card.background(color)

    def texts(
            self,
            texts: str | Sequence[str],
            layouts: str | Sequence[str],
            fonts: FreeTypeFont | Sequence[FreeTypeFont] | None = None,
            colors: Color | Sequence[Color] | None = None,
        ):
        """
        Add text strings to each card in this deck. Arguments may be scalar
        or sequences, varying the data per-card.
        """
        logger.debug(f"Deck.texts(sequence of strings)")
        args = self._make_seqs(texts=texts, layouts=layouts, fonts=fonts, colors=colors)
        for card, text, layout, font, color in zip(self._cards, args['texts'], args['layouts'], args['fonts'], args['colors']):
            card.text(text, layout, font, color)

    def images(
        self,
        images: Path | Sequence[Path],
        layouts: str | Sequence[str],
    ):
        """
        Draw external images on each card in this deck. Arguments may be scalar
        or sequences, varying the data per-card.
        """
        logger.debug(f"Deck.images(sequence of images)")
        args = self._make_seqs(images=images, layouts=layouts)
        for card, image, layout in zip(self._cards, args['images'], args['layouts']):
            card.image(image, layout)
    
    def _make_seqs(self, **kwargs):
        args = {}
        for key, value in kwargs.items():
            if not issubclass(type(value), Sequence) or issubclass(type(value), str):
                value = [value]
            args[key] = cycle(value)
        return args
    #endregion

    #region MutableSequence
    def __getitem__(self, index) -> Card:
        return self._cards[index]
    
    def __delitem__(self, index) -> None:
        del self._cards[index]
    
    def __iter__(self) -> Iterator[Card]:
        return self._cards.__iter__()

    def __len__(self) -> int:
        return len(self._cards)
    
    def __setitem__(self, index, value: Card) -> None:
        assert issubclass(type(value), Card)
        
        self._cards[index] = value
    
    def insert(self, index, value: Card) -> None:
        assert issubclass(type(value), Card)

        self._cards.insert(index, value)
    #endregion

    def __str__(self):
        return ", ".join([f"{idx}: {c}" for idx, c in enumerate(self._cards)])



from enum import Enum
from typing import NamedTuple, NewType
from numbers import Real
from PIL.ImageFont import FreeTypeFont

class Rect(NamedTuple):
    "A rectangle, the basic data structure used throughout Skit."
    x: Real
    """X coordinate for start of rectangle."""
    y: Real
    """Y coordinate for start of rectangle."""
    width: Real
    """Width of rectangle."""
    height: Real
    """Height of rectangle."""


class Alignment(Enum):
    "Where to align items in a layout."
    BEGIN = 'begin'
    """Align left if horizontal or top if vertical."""
    MIDDLE = 'middle'
    """Align to the middle."""
    END = 'end'
    """Align right if horitzontal or bottom if vertical."""


class Scale(Enum):
    "How to scale images within their layout."
    FIT = 'fit'
    """Scale the image up (larger) or down (smaller) to fit the layout."""
    DOWN = 'down'
    """Only scale the image down, never up."""
    UP = 'up'
    """Only scale the image up, never down."""
    NONE = 'none'
    """Disable scaling."""


class LayoutDef(NamedTuple):
    "A box for drawing into plus alignment information for the box's contents."
    x: Real
    """X coordinate for start of layout."""
    y: Real
    """Y coordinate for start of layout."""
    width: Real
    """Width of layout."""
    height: Real
    """Height of layout."""
    h_align: Alignment = Alignment.BEGIN
    """Specify horizontal alignment."""
    v_align: Alignment = Alignment.BEGIN
    """Specify vertical alignment."""
    scale: Scale = Scale.FIT
    """Specify how (and if) images are scaled."""


# color - thin wrapper on str right now, and there are only
# some forms that PIL allows, so this should be smarter one day
# https://pillow.readthedocs.io/en/stable/reference/ImageColor.html
Color = NewType('Color', str)

__all__ = [
    'Real',
    'FreeTypeFont',
    'Rect',
    'Alignment',
    'LayoutDef',
    'Color',
]


from collections.abc import Sequence
import logging
from pathlib import Path
from typing import Mapping
from PIL import Image
from skit._types import Real, LayoutDef, Color, FreeTypeFont, Alignment
from skit.render import DrawCommand, SingleImageRenderer
from abc import ABC, abstractmethod


logger = logging.getLogger(__file__)


class CardManipulation(ABC):
    @abstractmethod
    def background(self, color: str): pass

    @abstractmethod
    def layout(self, name: str, layoutdef: LayoutDef): pass

    @abstractmethod
    def layouts(self, names: Sequence[str], layoutdefs: Sequence[LayoutDef]): pass

    @abstractmethod
    def layouts_map(self, layouts: Mapping[str, LayoutDef]): pass

    @abstractmethod
    def text(
        self, text: str,
        layout: str,
        font: FreeTypeFont | None,
        color: Color | None,
    ): pass

    @abstractmethod
    def rectangle(
        self,
        layout: str,
        color: Color | None,
        thickness: Real | None,
    ): pass

    @abstractmethod
    def filled_rectangle(
        self,
        layout: str,
        color: Color,
    ): pass

    @abstractmethod
    def image(
        self,
        image: Path,
        layout: str,
    ): pass

    @abstractmethod
    def render_png(self, filename: str): pass

    @abstractmethod
    def render_pdf(self, filename: str, resolution: int, single_file: bool): pass


class Card(CardManipulation):
    "A single card."
    def __init__(self, width: int = 750, height: int = 1050):
        """Create a new card of a certain `width` and `height`."""
        self._width = width
        self._height = height
        self._layouts = {}
        self._background = '#ffffff00'
        self._commands = []

    def background(self, color: str):
        "Set the background color for this card."
        assert type(color) is str

        logger.debug(f"setting background to {color}")
        self._background = color

    def layout(self, name: str, layoutdef: LayoutDef):
        "Create a new layout for this card."
        assert Alignment(layoutdef.h_align)
        assert Alignment(layoutdef.v_align)

        logger.debug(f"creating layout area {name}")
        self._layouts[name] = {
            'x': layoutdef.x,
            'y': layoutdef.y,
            'width': layoutdef.width,
            'height': layoutdef.height,
            'h_align': layoutdef.h_align,
            'v_align': layoutdef.v_align,
            'scale': layoutdef.scale,
        }
    
    def layouts(self, names: Sequence[str], layoutdefs: Sequence[LayoutDef]):
        "Create multiple layouts for this card."
        assert len(names) == len(layoutdefs), "mismatched names/layoutdefs arguments"
        
        for name, layoutdef in zip(names, layoutdefs):
            self.layout(name, layoutdef)

    def layouts_map(self, layouts: Mapping[str, LayoutDef]):
        "Create multiple layouts from a dictionary."
        for name, layoutdef in layouts.items():
            self.layout(name, layoutdef)

    def text(self, text: str, layout: str, font: FreeTypeFont | None = None, color: Color | None = None):
        "Add text to this card."
        assert type(text) is str

        if layout in self._layouts:
            logger.debug(f"adding '{text}' in {layout}")
            self._commands.append({
                'op': DrawCommand.TEXT,
                'layout': layout,
                'text': text,
                'font': font,
                'color': color,
            })
        else:
            raise KeyError(f"missing layout '{layout}'")

    def rectangle(self, layout: str, color: Color | None = None, thickness: Real | None = None):
        "Draw a rectangle on this card."
        if layout in self._layouts:
            logger.debug(f"adding rectangle for {layout}")
            self._commands.append({
                'op': DrawCommand.RECTANGLE,
                'layout': layout,
                'color': color,
                'thickness': thickness,
                'filled': False,
            })
        else:
            raise KeyError(f"missing layout '{layout}'")

    def filled_rectangle(self, layout: str, color: Color):
        "Draw a filled rectangle on this card."
        if layout in self._layouts:
            logger.debug(f"adding rectangle for {layout}")
            self._commands.append({
                'op': DrawCommand.RECTANGLE,
                'layout': layout,
                'color': color,
                'thickness': 1,
                'filled': True,
            })
        else:
            raise KeyError(f"missing layout '{layout}'")

    def image(self, image: Path, layout: str):
        "Draw an external image on this card."
        if layout in self._layouts:
            logger.debug(f"adding image for {layout}")
            self._commands.append({
                'op': DrawCommand.IMAGE,
                'layout': layout,
                'image': image,
            })
        else:
            raise KeyError(f"missing layout '{layout}'")

    def render_png(self, filename: str):
        "Render this card as a PNG."
        logger.debug(f"rendering {filename}")

        im = (
            SingleImageRenderer(self._layouts)
            .render(self._width, self._height, self._background, self._commands)
        )
        im.save(filename, format='PNG')

    def render_pdf(self, filename: str, resolution: int, single_file=True):
        """
        Render this card as a PDF. It makes no sense to render a single
        card to multiple files, so `single_file` is ignored. PDFs don't
        support the alpha channel, so remove it.
        """
        logger.debug(f"rendering {filename}")

        im = self._get_rgb_image_for_pdf(resolution)
        im.save(filename, format='PDF', resolution=resolution)

    def _get_rgb_image_for_pdf(self, resolution: int):
        logging.debug(f"rendering RGB image")

        im = (
            SingleImageRenderer(self._layouts)
            .render(self._width, self._height, self._background, self._commands)
        )
        final = Image.new('RGB', im.size, self._background)
        final.paste(im)
        return final



"""
.. include:: ../README.md
"""
import copy
import warnings
from .deck import Deck
from .card import Card
from ._types import Rect, Color, Alignment, Scale, LayoutDef

from PIL import ImageFont


# expose ImageFont.truetype() as skit.load_font()
load_font = copy.copy(ImageFont.truetype)
load_font.__doc__ = f"""
A rename/import of [PIL.ImageFont.truetype][truetypedocs].

Required arguments are:
- `font`: A filename or file-like object containing a TrueType font.
If the file is not found in this filename, the loader may also search in other directories, such as the `fonts/`
directory on Windows or `/Library/Fonts/`, `/System/Library/Fonts/` and `~/Library/Fonts/` on macOS.
- `size`: The requested size, in pixels.

See the [upstream documentation][truetypedocs] for the remaining optional parameters.

[truetypedocs]: https://pillow.readthedocs.io/en/stable/reference/ImageFont.html#PIL.ImageFont.truetype
"""


def as_layoutdef(incoming_dict: dict) -> dict | LayoutDef:
    """
    A helper for converting dicts into `Rect`s. Useful parsing layouts from data
    files.

    Pass this to the `object_hook` parameter of `json.load()` to automatically
    convert dictionaries with the correct keys  into `LayoutDef`s.
    Correct keys are `x`, `y`, `width`, and `height`, plus optional
    `h_align`, `v_align`, and `scale`.

    Example:

    ```python
    with open('my_layout.json') as json_in:
        layouts = json.load(json_in, object_hook=skit.json_layout_hook)
    
    # assuming my_layout.json looked like this:
    # {
    #   "mybox": {
    #     "x": 0,
    #     "y": 0,
    #     "width": 100,
    #     "height": 25
    # }

    # then layouts will be:
    layouts = {
        'mybox': LayoutDef(
            x=0, y=0, width=100, height=25,
            h_align=Alignment.BEGIN, v_align=Alignment.BEGIN,
            scale=Scale.FIT,
        ),
    }

    ```
    """
    match incoming_dict:
        case { 'x': _, 'y': _, 'width': _, 'height': _, **rest }:
            if 'h_align' in rest:
                rest['h_align'] = Alignment(incoming_dict['h_align'])
            if 'v_align' in rest:
                rest['v_align'] = Alignment(incoming_dict['v_align'])
            if 'scale' in rest:
                rest['scale'] = Scale(incoming_dict['scale'])
            try:
                vals = incoming_dict.copy()
                vals.update(rest)
                return LayoutDef(**vals)
            except TypeError:
                # If there was something in the incoming data that doesn't
                # belong in LayoutDef, we'll hit this branch. We simply return
                # the dictionary we would have otherwise seen
                warnings.warn(f"{incoming_dict} has keys not convertible to LayoutDef")
                return incoming_dict
        case _:
            return incoming_dict


__all__ = [
    'Deck',
    'Card',
    'Rect',
    'Color',
    'Alignment',
    'Scale',
    'LayoutDef',
    'load_font',
    'as_layoutdef',
]
